ridgeLM <- function(Y, X, lambda, target=rep(0, ncol(X))){
	#######################################################################
	# Function evaluates the targeted ridge estimator of the linear 
	# regression model
	# 
	# Arguments
	# -> Y      : response vector
	# -> X      : design matrix
	# -> lambda : penalty parameter
	# -> target : target for regression parameter
	#
	# Author : Wessel N. van Wieringen
	#######################################################################

	if (ncol(X) == 1){
		return(solve(crossprod(X, X) + rep(lambda, ncol(X))) %*% 
		       (crossprod(X, Y) + lambda * target))
	} else {
		return(solve(crossprod(X, X) + diag(rep(lambda, ncol(X)))) %*% 
		       (crossprod(X, Y) + lambda * target))
	}
}

ssLM <- function(Y, X, betas){
	#######################################################################
	# Function calculates the sum of squares of a linear model
	# 
	# Arguments
	# -> Y      : response vector
	# -> X      : design matrix
	# -> betas  : regression parameter
	#
	# Author : Wessel N. van Wieringen
	#######################################################################

	error     <- Y - X %*% betas
	return(sum(error^2))
}

.kcvLMloss <- function(lambda, Y, X, folds, target){
	#######################################################################
	# Function yields the cross-validated sum-of-squares for the ridge
	# regression estimate with a nonnull target
	#
	# Arguments
	# -> lambda : penalty parameter
	# -> Y      : response vector
	# -> X      : design matrix
	# -> folds  : list with folds
	# -> target : target for regression parameter
	#
	# Author : Wessel N. van Wieringen
	#######################################################################

	# evaluate loss per fold
	cvLoss <- 0
	for (k in 1:length(folds)){
		cvLoss <- cvLoss + ssLM(Y[folds[[k]]], 
                                         X[folds[[k]],,drop=FALSE], 
                                         ridgeLM(Y[-folds[[k]]], 
	                                         X[-folds[[k]],,drop=FALSE], 
                                                 lambda, target))
	}
	
	# average over the folds
	return(cvLoss / length(folds))
}

.kcvLMpandp <- function(lambda, Y, X, folds, target, Ypast, Xpast){
	#######################################################################
	# Function yields the possibly modified cross-validated sum-of-squares 
	# for the targeted ridge regression estimate. The modification aims to 
	# discourage penalty parameter values that produce a targeted ridge 
	# estimator that performs poorly on historic data. The modification 
	# depends on the difference of the cross-validated sum-of-squares of 
	# the targeted ridge estimate and that of the target itself. If this 
	# difference exceeds a tolerated maximum and is in favor of the target, 
	# the cross-validated sum-of-squares is set equal to a very large value 
	# (10^10). If this difference is small but still in favor of the 
	# target, a barrier function is applied that multiplied the 
	# cross-validated sum-of-squares by a constant.The constant depends on
	# the difference itself and the sample sizes of historic and current 
	# data. The constant is one if the difference is zero. From one the 
	# constant monotonically increases with the difference, only to explode
	# at a preset the maximum tolerated difference. The modification is the
	# identity otherwise. 
	#
	# Arguments
	# -> lambda : penalty parameter
	# -> Y      : response vector of current data
	# -> X      : design matrix of current data
	# -> folds  : list with folds
	# -> target : target for regression parameter
	# -> Ypast  : response vector of historic data
	# -> Xpast  : design matrix of historic data
	#
	# Author : Wessel N. van Wieringen
	###########################################################

	# calculate loss of the target on historic data
	wssPastOld <- ssLM(Ypast, Xpast, target)

	# set storage variables
	cvLoss     <- 0
	newFitPast <- 0
	wssPastNew <- 0
	for (k in 1:length(folds)){
		# evaluate the targeted ridge estimator
		bHatR  <- ridgeLM(Y[-folds[[k]]], 
	                          X[-folds[[k]],], 
                                  lambda, target)

		# evaluate the loss of the targeted ridge estimator
		# on current data
		cvLoss <- cvLoss + ssLM(Y[folds[[k]]], 
                                        X[folds[[k]],,drop=FALSE], 
                                        bHatR)

		# evaluate the loss of the targeted ridge estimator
		# on historic data
		wssPastNew <- wssPastNew + ssLM(Ypast, Xpast, bHatR)
	}

	# average the cross-validated sum-of-squares
	cvLoss     <- cvLoss     / length(folds)
	wssPastNew <- wssPastNew / length(folds)
	fracNew    <- length(Y)  / (length(Ypast) + length(Y))

	# apply the barrier function
	if (wssPastOld / (1-fracNew) < wssPastNew){
		cvLoss <- cvLoss * 10^(10) 
	}
	if ((wssPastOld / (1-fracNew) > wssPastNew) & (wssPastNew > wssPastOld)){
		cvLoss <- cvLoss * (1 + (wssPastNew^2 / ((wssPastOld/(1-fracNew))^2 - wssPastNew^2) - (1-fracNew)^2 / (1 - (1-fracNew)^2)))
	}

	return(cvLoss)
}



optPenaltyLMpandp.kCVauto <- function(Y, X, lambdaInit, 
                                      fold=nrow(Y), target, 
                                      Ypast, Xpast){
	#######################################################################
	# Function finds the optimal penalty parameter of the targeted ridge 
	# regression estimator. The optimum is defined as the minimizer of the 
	# cross-validated sum-of-squares associated with the estimator such 
	# that the corresponding estimator does not inflate the historical
	# sum-of-squares too much. 
	#
	# Arguments
	# -> Y          : response vector of current data
	# -> X          : design matrix of current data
	# -> lambdaInit : initial guess of the optimal penalty parameter
	# -> fold       : number of folds to used in the cross-validation
	# -> target     : target for regression parameter
	# -> Ypast      : response vector of historic data
	# -> Xpast      : design matrix of historic data
	#
	# Author : Wessel N. van Wieringen
	#######################################################################

	# create folds
        fold    <- max(min(ceiling(fold), nrow(X)), 2)
        fold    <- rep(1:fold, ceiling(nrow(X)/fold))[1:nrow(X)]
        shuffle <- sample(1:nrow(X), nrow(X))
        folds   <- split(shuffle, as.factor(fold))

	# search by the Brent algorithm
	optLambda <- optim(lambdaInit, .kcvLMpandp, Y=Y, X=X, folds=folds, 
	                   target=target, Ypast=Ypast, Xpast=Xpast, 
	                   method="Brent", lower=10^(-10), upper=10^10)$par
	return(optLambda)
}




optPenaltyLM.kCVauto <- function(Y, X, lambdaInit, fold=nrow(X), target){
	#######################################################################
	# Function finds the optimal penalty parameter of the targeted ridge 
	# regression estimator. The optimum is defined as the minimizer of the 
	# cross-validated sum-of-squares associated with the estimator. 
	#
	# Arguments
	# -> Y          : response vector of current data
	# -> X          : design matrix of current data
	# -> lambdaInit : initial guess of the optimal penalty parameter
	# -> fold       : number of folds to used in the cross-validation
	# -> target     : target for regression parameter
	# -> Ypast      : response vector of historic data
	# -> Xpast      : design matrix of historic data
	#
	# Author : Wessel N. van Wieringen
	#######################################################################

	# create folds
        fold    <- max(min(ceiling(fold), nrow(X)), 2)
        fold    <- rep(1:fold, ceiling(nrow(X)/fold))[1:nrow(X)]
        shuffle <- sample(1:nrow(X), nrow(X))
        folds   <- split(shuffle, as.factor(fold))

	# search by the Brent algorithm
	optLambda <- optim(lambdaInit, .kcvLMloss, Y=Y, X=X, folds=folds, 
	                   target=target, method="Brent", lower=10^(-10), 
	                   upper=10^(10))$par
	return(optLambda)
}


##################################################################################
# Multiple targets
##################################################################################


ridgeLMmultT <- function(Y, X, lambdas, targetMat){
	##############################################################################
	# Function evaluates the ridge regression estimator when multiple targets
	# 
	# Arguments
	# -> Y          : Response vector
	# -> X          : Design matrix, variables in columns
	# -> lambdas    : Vector of penalty parameters, one per target
	# -> targetMat : Matrix with targets for the regression parameter as columns
	#
	# Author : Wessel N. van Wieringen
	##############################################################################

	# create weighted average of targets
	mixedTarget <- rep(0, ncol(X))
	lambdaTotal <- sum(lambdas)
	for (k in 1:length(lambdas)){
		mixedTarget <- mixedTarget + (lambdas[k] / lambdaTotal) * targetMat[,k]
	}

	# evaluate the ridge regression estimator with multiple targets
	return(ridgeLM(Y, X, lambdaTotal, mixedTarget))
}


.kcvlRidgeLMmultT <- function(lambdas, Y, X, targetMat, folds){
	##############################################################################
	# Function calculates a cross-validated negative log-likelihood score
	# for the multiple target ridge estimator of the parameter of the linear
	# regresson model for a particular value of the penalty parameter vector
	#
	# Arguments
	# -> lambdas   : Vector of penalty parameters, one per target
	# -> Y         : Response vector
	# -> X         : Design matrix, variables in columns
	# -> targetMat : Matrix with targets for the regression parameter as columns
	# -> folds     : cross-validation sample splits
	#
	# Author : Wessel N. van Wieringen
	##############################################################################

	# create weighted average of targets
	mixedTarget <- rep(0, ncol(X))
	lambdaTotal <- sum(lambdas)
	for (k in 1:length(lambdas)){
		mixedTarget <- mixedTarget + (lambdas[k] / lambdaTotal) * targetMat[,k]
	}

	# cross-validated loss evaluation
	cvLoss <- 0
	for (f in 1:length(folds)){
		cvLoss <- cvLoss + ssLM(Y[folds[[f]]], 
                                         X[folds[[f]],,drop=FALSE], 
                                         ridgeLM(Y[-folds[[f]]], 
	                                         X[-folds[[f]],], 
                                                 lambdaTotal, mixedTarget))
	}
	return(cvLoss / length(folds))
}



optPenaltyLMmultT.kCVauto <- function(Y, X, lambdas, fold=nrow(Y), 
                                      targetMat){
	#######################################################################
	# Function finds the optimal penalty parameter of the multi targeted
	# ridge regression estimator. The optimum is defined as the minimizer
	# of the cross-validated sum-of-squares associated with the estimator. 
	#
	# Arguments
	# -> Y         : Response vector
	# -> X         : Design matrix, variables in columns
	# -> lambdas   : Vector of penalty parameters, one per target
	# -> folds     : cross-validation sample splits
	# -> targetMat : Matrix with targets for the regression parameter as 
	#                  columns
	#
	# Author : Wessel N. van Wieringen
	#######################################################################

	# create folds
        fold    <- max(min(ceiling(fold), nrow(X)), 2)
        fold    <- rep(1:fold, ceiling(nrow(X)/fold))[1:nrow(X)]
        shuffle <- sample(1:nrow(X), nrow(X))
        folds   <- split(shuffle, as.factor(fold))

	# search for optimal parameters
	lambdasOpt <- constrOptim(lambdas, 
				  .kcvlRidgeLMmultT, 
				  grad=NULL, 
			          Y=Y, 
			          X=X, 
				  targetMat=targetMat, 
				  folds=folds, 
				  ui=diag(length(lambdas)), 
				  ci=rep(0, length(lambdas)))$par
	return(lambdasOpt)
}

.kcvRidgeLMmultTpandp <- function(lambdas, Y, X, folds, targetMat, Ypast, Xpast){
	#######################################################################
	# Function yields the possibly modified cross-validated sum-of-squares 
	# for the targeted ridge regression estimate. The modification aims to 
	# discourage penalty parameter values that produce a targeted ridge 
	# estimator that performs poorly on historic data. The modification 
	# depends on the difference of the cross-validated sum-of-squares of 
	# the targeted ridge estimate and that of the target itself. If this 
	# difference exceeds a tolerated maximum and is in favor of the target, 
	# the cross-validated sum-of-squares is set equal to a very large value 
	# (10^10). If this difference is small but still in favor of the 
	# target, a barrier function is applied that multiplied the 
	# cross-validated sum-of-squares by a constant.The constant depends on
	# the difference itself and the sample sizes of historic and current 
	# data. The constant is one if the difference is zero. From one the 
	# constant monotonically increases with the difference, only to explode
	# at a preset the maximum tolerated difference. The modification is the
	# identity otherwise. 
	#
	# Arguments
	# -> lambda    : penalty parameter
	# -> Y         : response vector of current data
	# -> X         : design matrix of current data
	# -> folds     : list with folds
	# -> targetMat : Matrix with targets for the regression parameter as 
	#                columns
	# -> Ypast     : response vector of historic data
	# -> Xpast     : design matrix of historic data
	#
	# Author : Wessel N. van Wieringen
	###########################################################

	# create weighted average of targets
	mixedTarget <- rep(0, ncol(X))
	lambdaTotal <- sum(lambdas)
	for (k in 1:length(lambdas)){
		mixedTarget <- mixedTarget + (lambdas[k] / lambdaTotal) * targetMat[,k]
	}

	# calculate for individual targets on historic data
	wssPastOld <- rep(0, length(lambdas))
	for (k in 1:length(lambdas)){
		wssPastOld[k] <- wssPastOld[k] + ssLM(Ypast, Xpast, targetMat[,k])
	}
	wssPastOld <- min(wssPastOld)

	# set storage variables
	cvLoss     <- 0
	newFitPast <- 0
	wssPastNew <- 0
	for (k in 1:length(folds)){
		# evaluate the targeted ridge estimator
		bHatR  <- ridgeLMmultT(Y[-folds[[k]]], 
	                               X[-folds[[k]],], 
                                       lambdas, targetMat)

		# evaluate the loss of the targeted ridge estimator
		# on current data
		cvLoss <- cvLoss + ssLM(Y[folds[[k]]], 
                                        X[folds[[k]],,drop=FALSE], 
                                        bHatR)

		# evaluate the loss of the targeted ridge estimator 
		# on historic data
		wssPastNew <- wssPastNew + ssLM(Ypast, Xpast, bHatR)
	}

	# average the cross-validated sum-of-squares
	cvLoss     <- cvLoss     / length(folds)
	wssPastNew <- wssPastNew / length(folds)
	fracNew    <- length(Y)  / (length(Ypast) + length(Y))

	# apply the barrier function
	if (wssPastOld / (1-fracNew) < wssPastNew){
		cvLoss <- cvLoss * 10^(10) 
	}
	if ((wssPastOld / (1-fracNew) > wssPastNew) & (wssPastNew > wssPastOld)){
		cvLoss <- cvLoss * (1 + (wssPastNew^2 / ((wssPastOld/(1-fracNew))^2 - wssPastNew^2) - (1-fracNew)^2 / (1 - (1-fracNew)^2)))
	}

	return(cvLoss)
}

optPenaltyLMmultTpandp.kCVauto <- function(Y, X, lambdas, fold=nrow(Y), 
                                           targetMat, Ypast, Xpast){
	#######################################################################
	# Function finds the optimal penalty parameter of the multi targeted
	# ridge regression estimator. The optimum is defined as the minimize
	# of the cross-validated sum-of-squares associated with the estimator
	# duch that the corresponding estimator does not inflate the historical
	# sum-of-squares too much. 
	#
	# Arguments
	# -> Y          : Response vector
	# -> X          : Design matrix, variables in columns
	# -> lambdas    : Vector of penalty parameters, one per target
	# -> fold       : number of folds to used in the cross-validation
	# -> targetMat   : Matrix with targets for the regression parameter as 
	#                  columns
	# -> folds       : cross-validation sample splits
	# -> Ypast      : response vector of historic data
	# -> Xpast      : design matrix of historic data
	#
	# Author : Wessel N. van Wieringen
	#######################################################################

	# create folds
        fold    <- max(min(ceiling(fold), nrow(X)), 2)
        fold    <- rep(1:fold, ceiling(nrow(X)/fold))[1:nrow(X)]
        shuffle <- sample(1:nrow(X), nrow(X))
        folds   <- split(shuffle, as.factor(fold))

	# search for optimal parameters
	lambdasOpt <- constrOptim(lambdas, 
				  .kcvRidgeLMmultTpandp, 
				  grad=NULL, 
			          Y=Y, 
			          X=X, 
				  targetMat=targetMat, 
				  folds=folds, 
	                          Ypast=Ypast, 
	                          Xpast=Xpast, 
				  ui=diag(length(lambdas)), 
				  ci=rep(0, length(lambdas)))$par
	return(lambdasOpt)
}



