# preliminaries
rm(list=ls())
gc()

# load libraries
library(mvtnorm)
library(penalized)



################################################################################
# declare functions
################################################################################

ncRRloss <- function(bHat, Y, X, b0, lambdaS, lambdaL){
	# nonzero centered ridge loss
	return(as.numeric(crossprod(Y - X %*% bHat) + lambdaS * crossprod(bHat - lambdaL*b0)))
}

ssLoss <- function(bHat, Y, X){
	# nonzero centered ridge loss
	return(as.numeric(crossprod(Y - X %*% bHat)))
}

ncRRcv2 <- function(lambdaSL, Y, X, b0){
	# cross-validation for nonzero centered ridge regression penalty determination
	cvLoss <- 0
	for (i in 1:length(Y)){
		bHat <- solve(t(X[-i,]) %*% X[-i,] + lambdaSL[1] * diag(ncol(X))) %*% (t(X[-i,]) %*% Y[-i] + lambdaSL[1] * lambdaSL[2] * b0)
		cvLoss <- cvLoss + ssLoss(bHat, Y[i], X[i,,drop=FALSE])
	}
	return(cvLoss/length(Y))
}

zcRRcv <- function(lambdaS, Y, X){
	# cross-validation for nonzero centered ridge regression penalty determination
	cvLoss <- 0
	for (i in 1:length(Y)){
		bHat <- solve(t(X[-i,]) %*% X[-i,] + lambdaS * diag(ncol(X))) %*% (t(X[-i,]) %*% Y[-i])
		cvLoss <- cvLoss + ssLoss(bHat, Y[i], X[i,,drop=FALSE])
	}
	return(cvLoss/length(Y))
}

ncRRcv1 <- function(lambdaS, Y, X, b0){
	# cross-validation for nonzero centered ridge regression penalty determination
	cvLoss <- 0
	for (i in 1:length(Y)){
		bHat <- solve(t(X[-i,]) %*% X[-i,] + lambdaS[1] * diag(ncol(X))) %*% (t(X[-i,]) %*% Y[-i] + lambdaS[1] * b0)
		cvLoss <- cvLoss + ssLoss(bHat, Y[i], X[i,,drop=FALSE])
	}
	return(cvLoss/length(Y))
}

ncRRcv1all <- function(lambdaS, Yall, Xall, t, n, b0){
	# cross-validation for nonzero centered ridge regression penalty determination
	cvLoss <- 0
	Y=Yall[(1:n) + n*(t-1)]
	X=Xall[(1:n) + n*(t-1),]
	for (i in 1:length(Y)){
		bHat <- solve(t(X[-i,]) %*% X[-i,] + lambdaS[1] * diag(ncol(X))) %*% (t(X[-i,]) %*% Y[-i] + lambdaS[1] * b0)
		cvLoss <- cvLoss + ssLoss(bHat, c(Yall[1:(n*(t-1))], Y[i]), rbind(Xall[1:(n*(t-1)),], X[i,,drop=FALSE]))
	}
	return(cvLoss/length(Y))
}






################################################################################
# start play
################################################################################

# set sample size, dimension, and number data sets in the sequence
n      <- 50
p      <- 100
nTimes <- 50

# set true parameter (p=11)
betas <- sort(c(1:(p/2) / (p/5), 1:(p/2) / (-p/5)))

# create storage objects
RRresList     <- list()
RRresPerfList <- list()
MMresList     <- list()
RRAresList    <- list()


for (u in 1:100){
	print(u)

	# generate data
	Xall <- numeric()
	Yall <- numeric()
	for (t in 1:nTimes){
		X <- matrix(rnorm(length(betas)*n), nrow=n)
		if ((t %% 10) != 0){
			Y <- X %*% betas + rnorm(n, sd=1)
		} else {
			Y <- rnorm(n, sd=1)
		}
		Xall <- rbind(Xall, X) 
		Yall <- c(Yall, Y)
	}

	# initiate result storage
	MMres1 <- matrix(ncol=p+3, nrow=0)
	colnames(MMres1) <- c("time", paste("b", 1:length(betas), sep=""), "eta", "bias")
	for (t in 1:nTimes){
		if (n*t > p){
			eta <- 1
			for (w in 1:100){
				etaPrev <- eta
				COVslh <- matrix(0, p, 1)
				VARslh <- matrix(0, p, p)
				for (k in 1:t){
 					Xtemp <- Xall[(1:n) + n*(k-1),]
					COVslh <- COVslh + t(Xtemp) %*% solve(eta * Xtemp %*% t(Xtemp) + diag(n)) %*% Yall[(1:n) + n*(k-1)]
					VARslh <- VARslh + t(Xtemp) %*% solve(eta * Xtemp %*% t(Xtemp) + diag(n)) %*% Xtemp
				}
				bHat      <- solve(VARslh) %*% COVslh 
	
				sigmaEhat <- 0
				for (k in 1:t){
 					Xtemp       <- Xall[(1:n) + n*(k-1),] 
					residualSlh <- Yall[(1:n) + n*(k-1)] - Xtemp %*% bHat
 					sigmaEhat   <- sigmaEhat + t(residualSlh) %*% solve(eta * Xtemp %*% t(Xtemp) + diag(n)) %*% residualSlh
				}
				sigmaEhat <- sigmaEhat / (n*t)

				loglikMM <- function(theta, X, Y, beta, sigmaE, t, n){
					detSlh <- 0
					VARslh <- matrix(0, p, p)
					loglik <- 0
					for (k in 1:t){
		 				Xtemp       <- Xall[(1:n) + n*(k-1),]
						detSlh      <- as.numeric(determinant(theta * Xtemp %*% t(Xtemp) + diag(n))$modulus[1])
						residualSlh <- Yall[(1:n) + n*(k-1)] - Xtemp  %*% bHat
			 			loglik      <- loglik + t(residualSlh) %*% solve(theta * Xtemp %*% t(Xtemp) + diag(n)) %*% residualSlh / sigmaEhat + 
                                                                detSlh + n*log(sigmaEhat)
					}
					return(loglik)
				}
				eta <- optim(eta, loglikMM, lower=10^(-5), upper=10^10, method="Brent", Y=Yall, X=Xall, beta=bHat, sigmaE=sigmaEhat, t=t, n=n)$par
				if ((etaPrev - eta)^2 < 10^(-8)){ break }
			}
			MMres1 <- rbind(MMres1, c(t, bHat, eta, sum((betas - bHat)^2)))
		} else{ 
			MMres1 <- rbind(MMres1, c(t, rep(NA, p+2)))
		}
	}
	MMresList[[u]] <- MMres1

	# initiate result storage
	RRres1 <- matrix(ncol=p+3, nrow=0)
	colnames(RRres1) <- c("time", paste("b", 1:length(betas), sep=""), "lambdaS", "bias")
	bHat <- rep(0, p)
	for (t in 1:nTimes){
		Xtemp       <- Xall[(1:n) + n*(t-1),]
		if (t==1){
			optLambda <- optim(1, ncRRcv1, lower=10^(-5), upper=10^10, method="Brent", Y=Yall[1:n], X=Xall[1:n,], b0=bHat)$par
		} else {
			optLambda <- optim(1, ncRRcv1all, lower=10^(-5), upper=10^10, method="Brent", Yall=Yall, Xall=Xall, t=t, n=n, b0=bHat)$par
		}
		bHat <- solve(t(Xtemp) %*% Xtemp + optLambda * diag(ncol(Xtemp))) %*% (t(Xtemp) %*% Yall[(1:n) + n*(t-1)] + optLambda * bHat)	
		RRres1 <- rbind(RRres1, c(t, bHat, optLambda, sum((betas - bHat)^2)))
	}
	RRresList[[u]] <- RRres1

	# initiate result storage
	RRres1 <- matrix(ncol=p+3, nrow=0)
	colnames(RRres1) <- c("time", paste("b", 1:length(betas), sep=""), "lambdaS", "bias")
	bHat <- betas
	for (t in 1:nTimes){
		Xtemp       <- Xall[(1:n) + n*(t-1),]
		if (t==1){
			optLambda <- optim(1, ncRRcv1, lower=10^(-5), upper=10^10, method="Brent", Y=Yall[1:n], X=Xall[1:n,], b0=bHat)$par
		} else {
			optLambda <- optim(1, ncRRcv1all, lower=10^(-5), upper=10^10, method="Brent", Yall=Yall, Xall=Xall, t=t, n=n, b0=bHat)$par
		}
		bHat <- solve(t(Xtemp) %*% Xtemp + optLambda * diag(ncol(Xtemp))) %*% (t(Xtemp) %*% Yall[(1:n) + n*(t-1)] + optLambda * bHat)	
		RRres1 <- rbind(RRres1, c(t, bHat, optLambda, sum((betas - bHat)^2)))
	}
	RRresPerfList[[u]] <- RRres1

	# initiate result storage
	RRAres1           <- matrix(ncol=length(betas)+3, nrow=0)
	colnames(RRAres1) <- c("time", paste("b", 1:length(betas), sep=""), "eta", "bias")
	for (t in 1:nTimes){
		optLambda <- optPenaltyGLM.kCVauto(Y=Yall[1:(n*t)], X=Xall[1:(n*t),], 1, fold=10, target=0*betas, model="linear", stratified=TRUE)
		bHat      <- ridgeGLM(Y=Yall[1:(n*t)], X=Xall[1:(n*t),], lambda=optLambda, model="linear", target=0*betas)
		RRAres1   <- rbind(RRAres1, c(t, bHat, optLambda, sum((betas - bHat)^2)))
	}
	RRAresList[[u]] <- RRAres1
}

##################################################################################################################################

# plot results
ymax <- 0
ymin <- 0
for (k in 1:100){
	ymax <- max(ymax, max(RRresAltList[[k]][,104], MMresList[[k]][,104], RRAresList[[k]][,104], RRresPerfList[[k]][,104], na.rm=TRUE))
	ymin <- min(ymin, min(RRresAltList[[k]][,104], MMresList[[k]][,104], RRAresList[[k]][,104], RRresPerfList[[k]][,104], na.rm=TRUE))
}
ylims <- c(ymin, ymax)
plot(RRresList[[1]][,104], col="white", ylim=c(-5.4, log(ylims[2])), xlab="update", ylab="log(MSE)", main="MSE comparison of mixed model and ridge estimators")
for (k in 1:100){
	lines(log(RRresAltList[[k]][,104]), col="red", lty=2)	
	lines(log(MMresList[[k]][,104]), col="black", lty=1)
	lines(log(RRresPerfList[[k]][,104]), col="orange", lty=4)
	lines(log(RRAresList[[k]][,104]),   col="purple", lty=5)
}
legend("topright", c("mixed model",
	             expression(paste("update ridge (", beta[0], "=", 0[p], ")", sep="")),
        	     expression(paste("update ridge (", beta[0], "=", beta, ")", sep="")),
                     "ridge, all data"), 
		     lty=c(1, 2, 4, 5), col=c("black", "red", "orange", "purple"))





